function GCBackupManager()
{
	this.atom = Namespace("", "http://www.w3.org/2005/Atom");
	this.openSearch = Namespace("openSearch", "http://a9.com/-/spec/opensearch/1.1/");
	this.gContact = Namespace("gContact", "http://schemas.google.com/contact/2008");
	this.gd = Namespace("gd", "http://schemas.google.com/g/2005");
	this.batch =  Namespace("batch", "http://schemas.google.com/gdata/batch");
	this.cardFields =
        [ // Contact > Name
         ["First Name", "FirstName"],
         ["Last Name", "LastName"],
         ["Display Name", "DisplayName"],
         ["NickName", "NickName"],
          // Contact > Internet
         ["Primary Email", "PrimaryEmail"],
         ["Secondary Email", "SecondEmail"],
          // Contact > Phones
         ["Work Phone", "WorkPhone"],
         ["Home Phone", "HomePhone"],
         ["Fax Number", "FaxNumber"],
         ["Pager Number", "PagerNumber"],
         ["Mobile Number", "CellularNumber"],
          // Address > Home
         ["Home Address", "HomeAddress"],
         ["Home Address 2", "HomeAddress2"],
         ["Home City", "HomeCity"],
         ["Home State", "HomeState"],
         ["Home ZipCode", "HomeZipCode"],
         ["Home Country", "HomeCountry"],
          // Address > Work
         ["Work Address", "WorkAddress"],
         ["Work Address 2", "WorkAddress2"],
         ["Work City", "WorkCity"],
         ["Work State", "WorkState"],
         ["Work ZipCode", "WorkZipCode"],
         ["Work Country", "WorkCountry"],
         ["Job Title", "JobTitle"],
         ["Department", "Department"],
         ["Organization", "Company"],
         ["Web Page 1", "WebPage1"],
         ["Web Page 2", "WebPage2"],
         ["Birth Year", "BirthYear"],
         ["Birth Month", "BirthMonth"],
         ["Birth Day", "BirthDay"],
          // Other > (custom)
         ["Custom 1", "Custom1"],
         ["Custom 2", "Custom2"],
         ["Custom 3", "Custom3"],
         ["Custom 4", "Custom4"],
          // Other > Notes
         ["Notes", "Notes"]
         // Not Exported
         //"aimScreenName"
         //"PhoneticLastName"
         //"PhoneticFirstName"
         //"allowRemoteContent"
         //"Age"
         ];
	this.pref = Components.classes['@mozilla.org/preferences;1'].
	  						getService(Components.interfaces.nsIPrefBranch);
	this.localeBundle2 = document.getElementById("GmContBundle2");
	var date = new Date();
	this.backupDirName = date.getFullYear().toString()
					 					 + ("0"+(date.getMonth()+1)).substr(-2)
					 					 + ("0"+date.getDate()).substr(-2)
					 					 + "-"
					 					 + ("0"+date.getHours()).substr(-2)
					 					 + ("0"+date.getMinutes()).substr(-2)
					 					 + ("0"+date.getSeconds()).substr(-2)
					 					 + ("00"+date.getMilliseconds()).substr(-3);
}

GCBackupManager.prototype.getDumpDir = function(account)
{
	var file = null;
	try {
		var value = this.pref.getComplexValue("gmcont.dump_rel_path", Components.interfaces.nsIRelativeFilePref);
		file = value.file;
		if (!file || !file.isWritable() || !file.isReadable()) file = null;
	} catch(e) {
		file = null;
	}
	
	try {
		if (!file) {
			file = Components.classes["@mozilla.org/file/directory_service;1"]
								.getService(Components.interfaces.nsIProperties)
								.get("ProfD", Components.interfaces.nsILocalFile);
			file.append("GoogleContacts");
		}
		if (account) file.append(account.replace("@","_"));
	} catch(e) {
		dump("Error: Cannot get dump dir.\n");
		file = null;
	}
	
	return file;
}

GCBackupManager.prototype.getCacheDir = function()
{
	var file = this.getDumpDir();
	if (file) file.append("cache");
	
	return file;
}

GCBackupManager.prototype.getLast2Files = function(file)
{
	if (!file.exists()) return null;
	var list = new Array();
	var entries = file.directoryEntries;
	while (entries.hasMoreElements()) {
		var entry = entries.getNext().QueryInterface(Components.interfaces.nsILocalFile);
		if (!entry.isDirectory()) list.push(entry);
	}
	if (list.length == 0) return null;
	if (index != null && index >= list.length) return null;
	
	var sortFunc = function(a, b) {
		return a.leafName < b.leafName ? -1 : 1;
	};	
	
	list.sort(sortFunc);
	
	var ret = {
		first: list.pop(),
		second: list.pop()
	}
	
	return ret;
}

GCBackupManager.prototype.prepareDump = function(file, maxNum)
{
	if (!file.exists()) return;
	var list = new Array();
	var entries = file.directoryEntries;
	while (entries.hasMoreElements()) {
		var entry = entries.getNext().QueryInterface(Components.interfaces.nsILocalFile);
		//if (!entry.isDirectory()) list.push(entry);
		list.push(entry);
	}
	
	var sortFunc = function(a, b) {
		//return a.leafName < b.leafName ? -1 : 1;
		return a.lastModifiedTime < b.lastModifiedTime ? -1 : 1;
	};	
	
	list.sort(sortFunc);
	if (maxNum > -1) {
		var over = list.length - maxNum + 1;
		for (var i=0; i<over; i++) {
			if (list[i]) list[i].remove(true);
		}
	}
}

GCBackupManager.prototype.dump = function(file, account, str, ext, charset)
{
	file.append(this.backupDirName);
	var date = new Date();
	var name = account.replace("@","_")
					 + "-"
					 + date.getFullYear().toString()
					 + ("0"+(date.getMonth()+1)).substr(-2)
					 + ("0"+date.getDate()).substr(-2)
					 + "-"
					 + ("0"+date.getHours()).substr(-2)
					 + ("0"+date.getMinutes()).substr(-2)
					 + ("0"+date.getSeconds()).substr(-2)
					 + ("00"+date.getMilliseconds()).substr(-3)
					 + "." + ext;
	file.append(name);
	if (file.exists()) {
		file.remove(true);
	}
	file.create(file.NORMAL_FILE_TYPE, 0666);
	var ioService = Components.classes['@mozilla.org/network/io-service;1']
                    .getService(Components.interfaces.nsIIOService);

	var fileStream = Components.classes['@mozilla.org/network/file-output-stream;1']
										.createInstance(Components.interfaces.nsIFileOutputStream);
	fileStream.init(file, 2, 0x200, false);
	//use default charset
	//XXX: How about getting native encoding?
	var encoding = charset ? charset
												 : nsPreferences.copyUnicharPref(
												 									"gmcont.dump_charset",
												 									this.localeBundle2.getString("intl.charset.default")
												 									);
	//dump("Dump file: "+file.path+"\n");
	var converterStream;
	try {
		converterStream = Components
			.classes['@mozilla.org/intl/converter-output-stream;1']
			.createInstance(Components.interfaces.nsIConverterOutputStream);
		converterStream.init(fileStream, encoding, 0,
											 Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
		converterStream.writeString(str);
	} catch (e) {
		try {										 
			converterStream.init(fileStream, "UTF-8", 0,
												 Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
			converterStream.writeString(str);
		} catch(e) {
			dump(e);
		}
	}
	converterStream.close();
	fileStream.close();
}

GCBackupManager.prototype.diff = function(file1, file2)
{
	if (!file1 || !file2) return true;
	
  var stream1 = Components.classes['@mozilla.org/network/file-input-stream;1'].createInstance(Components.interfaces.nsIFileInputStream);
  var stream2 = Components.classes['@mozilla.org/network/file-input-stream;1'].createInstance(Components.interfaces.nsIFileInputStream);
  try {
	  stream1.init(file1, 1, 0, false);
	  stream2.init(file2, 1, 0, false);
    var scriptableStream1 = Components.classes['@mozilla.org/scriptableinputstream;1'].createInstance(Components.interfaces.nsIScriptableInputStream);
    var scriptableStream2 = Components.classes['@mozilla.org/scriptableinputstream;1'].createInstance(Components.interfaces.nsIScriptableInputStream);
    scriptableStream1.init(stream1);
    scriptableStream2.init(stream2);
    var fileSize1 = scriptableStream1.available();
    var fileSize2 = scriptableStream2.available();
    
    if (fileSize1 != fileSize2) return true;
    
    while(scriptableStream1.available()) {
    	var fileContents1 = scriptableStream1.read(512);	
    	var fileContents2 = scriptableStream2.read(512);
    	if (fileContents1 != fileContents2) {
 		    scriptableStream1.close();
		    stream1.close();
 		    scriptableStream2.close();
		    stream2.close();
    		return true;
    	}
    }

    scriptableStream1.close();
    stream1.close();
    scriptableStream2.close();
    stream2.close();
    return false; // same files
  } catch(e) {
    return true;
  }	
}

GCBackupManager.prototype.isUpdateNeeded = function(account, feedXML, isGroup)
{
	var feedStr = feedXML.toString();
	feedStr = feedStr.replace(/gd:etag=\S+/,"");
	feedStr = feedStr.replace(/\<updated\>\S+\<\/updated\>/,"");
	
	var file = this.getDumpDir(account);
	file.append("xml");
	if (isGroup) file.append("groups");
	else file.append("contacts");

	this.prepareDump(file, 2);
	this.dump(file, account, feedStr, "xml", "UTF-8");
	
	file = this.getDumpDir(account);
	file.append("xml");
	if (isGroup) file.append("groups");
	else file.append("contacts");

	var diffFiles = this.getLast2Files(file);
	return diffFiles ? this.diff(diffFiles.first, diffFiles.second) : true;
	//return true;
}

GCBackupManager.prototype.dumpRawXMLString = function(account, feedStr, isGroup, startIndex)
{
	//var feedStr = feedXML.toString();
	//feedStr = feedStr.replace(/gd:etag=\S+/,"");
	//feedStr = feedStr.replace(/\<updated\>\S+\<\/updated\>/,"");
	
	var file = this.getDumpDir(account);
	if (!file) return;
	file.append("xml");
	if (isGroup) file.append("groups");
	else file.append("contacts");

	//this.prepareDump(file, 1);
	if (startIndex == 1) this.prepareDump(file, 1);
	else this.prepareDump(file, -1);
	this.dump(file, account, feedStr, "xml", "UTF-8");
}

GCBackupManager.prototype.dumpGoogleContacts = function(account, feedXML)
{
	var maxNum = nsPreferences.getIntPref("gmcont.max_backups", 0);
	if (maxNum <= 0) return;
	
	var file = this.getDumpDir(account);
	if (!file) return;
	
	file.append("Google");
	this.prepareDump(file, maxNum);
	var feedStr = feedXML.toString();
	var csvStr = this.feedToCSV(feedXML);
	var charset = nsPreferences.copyUnicharPref("gmcont.dump_charset_gc", "UTF-16");
	this.dump(file, account, csvStr, "csv", charset);
}

GCBackupManager.prototype.copyTBAddressBook = function(account, abFile)
{
	var maxNum = nsPreferences.getIntPref("gmcont.max_backups", 0);
	if (maxNum <= 0) return;

	var file = this.getDumpDir(account);
	if (!file) return;
	
	file.append("Thunderbird");
	this.prepareDump(file, maxNum);	
	file.append(this.backupDirName);
	if (file.exists()) {
		file.remove(true);
	}
	/*
	file.append(name);
	if (file.exists()) {
		file.remove(true);
	}
	*/
	var name = abFile.leafName;
	abFile.copyTo(file, name);
}

GCBackupManager.prototype.dumpTBAddressBook = function(account, cards)
{
	var maxNum = nsPreferences.getIntPref("gmcont.max_backups", 0);
	if (maxNum <= 0) return;
	
	var file = this.getDumpDir(account);
	if (!file) return;
	
	file.append("Thunderbird");
	this.prepareDump(file, maxNum);
	
	var dumpStr = "";
	var cols = new Array();
	for (var i=0; i<this.cardFields.length; i++) {
		cols.push(this.cardFields[i][0]);
	}
	dumpStr += cols.join(",") + "\n";
	
	for (var i=0; i<cards.length; i++) {
		var card = cards[i];
		var row = new Array();
		for (var j=0; j<this.cardFields.length; j++) {
			//var data = card[this.cardFields[j][1]];
			var data = card.getProperty(this.cardFields[j][1], null);
			if (!data) data = "";
//			data = data.replace(/"/g,'""');
//			if (data.indexOf('"') >=0 || data.indexOf(",") >= 0) data = '"' + data + '"';
			data = this.escapeForCSV(data);
			row.push(data);
		}
		dumpStr += row.join(",") + "\n";
	}
	var charset = nsPreferences.copyUnicharPref("gmcont.dump_charset", null);
	this.dump(file, account, dumpStr, "csv", charset);	
}

GCBackupManager.prototype.escapeForCSV = function(data)
{
	data = data.toString();
	data = data.replace(/"/g,'""');
	if (data.indexOf('"') >=0 || data.indexOf(",") >= 0 || data.indexOf("\n") >= 0) data = '"' + data + '"';
	return data;
}
/*
GCBackupManager.prototype.feedV3ToCSV = function(feedXML)
{
	var atom = this.atom;
	var gd = this.gd;
	var gContact = this.gContact;
	var contactsXML = new XML(feedXML);
	var colTemplate = [
		"Name",
		"Given Name",
		"Additional Name",
		"Family Name",
		"Yomi Name",
		"Given Name Yomi",
		"Additional Name Yomi",
		"Family Name Yomi",
		"Name Prefix",
		"Name Suffix",
		"Initials",
		"Nickname",
		"Short Name",
		"Maiden Name",
		"Birthday",
		"Gender",
		"Location",
		"Billing Information",
		"Directory Server",
		"Mileage",
		"Occupation",
		"Hobby",
		"Sensitivity",
		"Priority",
		"Subject",
		"Notes",
		"Group Membership",
		"E-mail %d - Type",
		"E-mail %d - Value",
		"IM %d - Type",
		"IM %d - Service",
		"IM %d - Value",
		"Phone %d - Type",
		"Phone %d - Value",
		"Address %d - Type",
		"Address %d - Formatted",
		"Address %d - Street",
		"Address %d - City",
		"Address %d - PO Box",
		Address 1 - Region,Address 1 - Postal Code,Address 1 - Country,Address 1 - Extended Address,Address 2 - Type,Address 2 - Formatted,Address 2 - Street,Address 2 - City,Address 2 - PO Box,Address 2 - Region,Address 2 - Postal Code,Address 2 - Country,Address 2 - Extended Address,Organization 1 - Type,Organization 1 - Name,Organization 1 - Yomi Name,Organization 1 - Title,Organization 1 - Department,Organization 1 - Symbol,Organization 1 - Location,Organization 1 - Job Description,Website 1 - Type,Website 1 - Value,Website 2 - Type,Website 2 - Value,Custom Field 1 - Type,Custom Field 1 - Value,Custom Field 2 - Type,Custom Field 2 - Value,Custom Field 3 - Type,Custom Field 3 - Value
	];
	var entries = contactsXML.atom::entry;
	
}
*/
GCBackupManager.prototype.feedToCSV = function(feedXML)
{	
	var atom = this.atom;
	var gd = this.gd;
	var gContact = this.gContact;
	var csvStr = "Name,E-mail,Notes,Section 1 - Description,Section 1 - Email,Section 1 - IM,Section 1 - Phone,Section 1 - Mobile,Section 1 - Pager,Section 1 - Fax,Section 1 - Company,Section 1 - Title,Section 1 - Other,Section 1 - Address,Section 2 - Description,Section 2 - Email,Section 2 - IM,Section 2 - Phone,Section 2 - Mobile,Section 2 - Pager,Section 2 - Fax,Section 2 - Company,Section 2 - Title,Section 2 - Other,Section 2 - Address,Section 3 - Description,Section 3 - Email,Section 3 - IM,Section 3 - Phone,Section 3 - Mobile,Section 3 - Pager,Section 3 - Fax,Section 3 - Company,Section 3 - Title,Section 3 - Other,Section 3 - Address,Section 4 - Description,Section 4 - Email,Section 4 - IM,Section 4 - Phone,Section 4 - Mobile,Section 4 - Pager,Section 4 - Fax,Section 4 - Company,Section 4 - Title,Section 4 - Other,Section 4 - Address\n";
	var contactsXML = new XML(feedXML);
	var entries = contactsXML.atom::entry;
	for (var i=0; i<entries.length(); i++) {
		var entry = entries[i];
		//Name
		var data = this.parseTitle(entry);
		csvStr += this.escapeForCSV(data);
			
		var personalSection = {
			description: "Personal",
			email: [],
			im: [],
			phone: [],
			mobile: [],
			pager: [],
			fax: [],
			company: [],
			title: [],
			other: [],
			address: []
		};
		var workSection = {
			description: "Work",
			email: [],
			im: [],
			phone: [],
			mobile: [],
			pager: [],
			fax: [],
			company: [],
			title: [],
			other: [],
			address: []
		};
		var otherSection = {
			description: "Other",
			email: [],
			im: [],
			phone: [],
			mobile: [],
			pager: [],
			fax: [],
			company: [],
			title: [],
			other: [],
			address: []
		};
		var customSection = {
			description: "CUSTOM",
			email: [],
			im: [],
			phone: [],
			mobile: [],
			pager: [],
			fax: [],
			company: [],
			title: [],
			other: [],
			address: []
		};
		var email = this.parseEmail(entry, personalSection, workSection, otherSection, customSection);
		this.parseIM(entry, personalSection, workSection, otherSection, customSection);
		this.parsePhone(entry, personalSection, workSection, otherSection, customSection);
		workSection.company.push(this.parseOrgName(entry, personalSection, workSection, otherSection, customSection));
		workSection.title.push(this.parseOrgTitle(entry, personalSection, workSection, otherSection, customSection));
		this.parsePostalAddress(entry, personalSection, workSection, otherSection, customSection);
		csvStr += "," + email;
		data = this.parseContent(entry);
		csvStr += "," + this.escapeForCSV(data);
		csvStr += "," + this.sectionToCSV(personalSection);
		csvStr += "," + this.sectionToCSV(workSection);
		csvStr += "," + this.sectionToCSV(otherSection);
		csvStr += "," + this.sectionToCSV(customSection) + "\n";
	}
	
	return csvStr;
}

GCBackupManager.prototype.sectionToCSV = function(section)
{
	var str = section.description + ","
					+ section.email.join("; ") + ","
					+ section.im.join("; ") + ","
					+ section.phone.join("; ") + ","
					+ section.mobile.join("; ") + ","
					+ section.pager.join("; ") + ","
					+ section.fax.join("; ") + ","
					+ section.company.join("; ") + ","
					+ section.title.join("; ") + ","
					+ section.other.join("; ") + ","
					+ section.address.join("; ");
	return str;
}

GCBackupManager.prototype.parseTitle = function(entry)
{
	var atom = this.atom;
	var title = "";
	try {
		title = entry.atom::title;
	} catch(e) {}	
	return this.escapeForCSV(title);
}

GCBackupManager.prototype.parseContent = function(entry)
{
	var atom = this.atom;
	var content = "";
	try {
		content = entry.atom::content;
	} catch(e) {}	
	return this.escapeForCSV(content);
}

GCBackupManager.prototype.parseOrgTitle = function(entry)
{
	var atom = this.atom;
	var title = "";
	try {
		title = entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work")[0].gd::orgTitle;
	} catch(e) {
	}
	return this.escapeForCSV(title);
}

GCBackupManager.prototype.parseOrgName = function(entry)
{
	var atom = this.atom;
	var name = "";
	try {
		name = entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work")[0].gd::orgName;
	} catch(e) {
	}
	return this.escapeForCSV(name);
}

GCBackupManager.prototype.parseEmail = function(entry, personal, work, other, custom)
{
	var gd = this.gd;
	var emails = entry.gd::email;
	var primaryEmail = "";
	try {
		primaryEmail = entry.gd::email[0].@address;
	} catch(e) {
	}
	
	for (var i=1; i<emails.length(); i++) { //skip primary email
		var email = emails[i];
		var desc = email.@rel.toString();
		if (desc == "http://schemas.google.com/g/2005#home") {
			personal.email.push(email.@address);
		} else if (desc == "http://schemas.google.com/g/2005#work") {
			work.email.push(email.@address);
		} else if (desc == "http://schemas.google.com/g/2005#other") {
			other.email.push(email.@address);
		}
	}
	
	return primaryEmail;
}

GCBackupManager.prototype.parseIM = function(entry, personal, work, other, custom)
{
	var gd = this.gd;
	var ims = entry.gd::im;
	
	for (var i=0; i<ims.length(); i++) {
		var im = ims[i];
		var label = im.@label.toString();
		var protocol = im.@protocol.toString();
		var app = "";
		switch (protocol) {
			case "http://schemas.google.com/g/2005#GOOGLE_TALK":
				app = "GTALK";
				break;
			case "http://schemas.google.com/g/2005#AIM":
				app = "AIM";
				break;
			case "http://schemas.google.com/g/2005#YAHOO":
				app = "YAHOO";
				break;
			case "http://schemas.google.com/g/2005#MSN":
				app = "MSN";
				break;
			case "http://schemas.google.com/g/2005#JABBER":
				app = "JABBER";
				break;
			case "http://schemas.google.com/g/2005#ICQ":
				app = "ICQ";
				break;
		}

		if (label == "CUSTOM") { //it seems that label is CUSTOM only so far
			custom.im.push(app + ": "+this.escapeForCSV(im.@address));
		} else if (label == "Personal") {
			personal.im.push(app + ": "+this.escapeForCSV(im.@address));
		} else if (label == "Work") {
			work.im.push(app + ": "+this.escapeForCSV(im.@address));
		} else if (label == "Other") {
			other.im.push(app + ": "+this.escapeForCSV(im.@address));
		}
	}
}

GCBackupManager.prototype.parsePhone = function(entry, personal, work, other, custom)
{
	var gd = this.gd;
	var phones = entry.gd::phoneNumber;
	
	for (var i=0; i<phones.length(); i++) {
		var phone = phones[i];
		var desc = phone.@rel.toString();
		switch (desc) {
			case "http://schemas.google.com/g/2005#mobile":
				personal.mobile.push(this.escapeForCSV(phone));
				break;
			case "http://schemas.google.com/g/2005#home":
				personal.phone.push(this.escapeForCSV(phone));
				break;
			case "http://schemas.google.com/g/2005#work":
				work.phone.push(this.escapeForCSV(phone));
				break;
			case "http://schemas.google.com/g/2005#home_fax":
				personal.fax.push(this.escapeForCSV(phone));
				break;
			case "http://schemas.google.com/g/2005#work_fax":
				work.fax.push(this.escapeForCSV(phone));
				break;
			case "http://schemas.google.com/g/2005#pager":
				personal.pager.push(this.escapeForCSV(phone));
				break;
			case "http://schemas.google.com/g/2005#other":
				other.fax.push(this.escapeForCSV(phone));
				break;
		}
	}	
}

GCBackupManager.prototype.parsePostalAddress = function(entry, personal, work, other, custom)
{
	var gd = this.gd;
	var addrs = entry.gd::postalAddress;
	
	for (var i=0; i<addrs.length(); i++) {
		var addr = addrs[i];
		var desc = addr.@rel.toString();
		switch (desc) {
			case "http://schemas.google.com/g/2005#home":
				personal.address.push(this.escapeForCSV(addr));
				break;
			case "http://schemas.google.com/g/2005#work":
				work.address.push(this.escapeForCSV(addr));
				break;
			case "http://schemas.google.com/g/2005#other":
				other.address.push(this.escapeForCSV(addr));
				break;
		}
	}	
}
