//Super class for contact and group
function GCXMLEntry(entry)
{
	this.atom = Namespace("", "http://www.w3.org/2005/Atom");
	this.openSearch = Namespace("openSearch", "http://a9.com/-/spec/opensearch/1.1/");
	this.gContact = Namespace("gContact", "http://schemas.google.com/contact/2008");
	this.gd = Namespace("gd", "http://schemas.google.com/g/2005");
	this.batch =  Namespace("batch", "http://schemas.google.com/gdata/batch");
	this.localeBundle = document.getElementById("GmContBundle");
	this.itemObj = {};
	this.useDiskCache = nsPreferences.getBoolPref("gmcont.use_disk_cache", true);
	this.entry = entry;
	this.cacheFileName = null;
	this.saveCache();
}

GCXMLEntry.prototype.getCacheDir = function()
{
	var backupMgr = new GCBackupManager();
	var file = backupMgr.getCacheDir();
	
	return file;
}

GCXMLEntry.prototype.loadCache = function()
{
	if (!this.useDiskCache || !this.cacheFileName) return false;
	var file = this.getCacheDir();
	if (!file) return false;
	file.append(this.cacheFileName);
	if (!file.exists()) return false;

	var stream = Components.classes['@mozilla.org/network/file-input-stream;1'].createInstance(Components.interfaces.nsIFileInputStream);
  try {
	  stream.init(file, 1, 0, false);
  	var converterStream = Components
				.classes['@mozilla.org/intl/converter-input-stream;1']
				.createInstance(Components.interfaces.nsIConverterInputStream);
		converterStream.init(stream, "UTF-8", stream.available(),
												 Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
		var cacheObj = {};
		converterStream.readString(stream.available(), cacheObj);
		converterStream.close();
		stream.close();
		this.entry = new XML(cacheObj.value);
	} catch(e) {
		dump(e);
		return false;
	}

	return true;
}

GCXMLEntry.prototype.saveCache = function()
{
	if (!this.entry || !this.useDiskCache) return false;

	var file = this.getCacheDir();
	this.cacheFileName = this.getId().toString();
	this.cacheFileName = this.cacheFileName.replace(/[@:/%]/g,"_");
	file.append(this.cacheFileName);
	
	if (file.exists()) file.remove(true);
	file.create(file.NORMAL_FILE_TYPE, 0666);

	var fileStream = Components.classes['@mozilla.org/network/file-output-stream;1']
										.createInstance(Components.interfaces.nsIFileOutputStream);
	fileStream.init(file, 2, 0x200, false);
	
	var converterStream = Components
			.classes['@mozilla.org/intl/converter-output-stream;1']
			.createInstance(Components.interfaces.nsIConverterOutputStream);
			
	try {										 
		converterStream.init(fileStream, "UTF-8", 0,
											 Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
		converterStream.writeString(this.entry.toString());
	} catch(e) {
		dump(e);
		return false;
	}

	converterStream.close();
	fileStream.close();
	
	return true;
}

GCXMLEntry.prototype.removeCache = function()
{
	if (!this.cacheFileName) return false;
	
	var file = this.getCacheDir();
	file.append(this.cacheFileName);
	if (file.exists()) file.remove(false);

	return true;
}

GCXMLEntry.prototype.purgeMemCache = function()
{
	if (!this.useDiskCache) return false;
	delete this.entry;
	this.entry = null;
	
	return true;
}

GCXMLEntry.prototype.toString = function()
{
	if (!this.entry && !this.loadCache()) return "";
	return this.entry.toString();
}

GCXMLEntry.prototype.toXMLString = function(rmEtag)
{
	var atom = this.atom;
	var gd = this.gd;
	if (!this.entry && !this.loadCache()) return "";
	delete this.entry.@gd::etag;
	var xmlString = this.entry.toXMLString(); 
	return xmlString;
}

GCXMLEntry.prototype.init = function(feed)
{
	var atom = this.atom;
	// remove <?xml ?>
	this.entry = new XML(feed.replace(/\<\?xml.+\?\>/,""));
	this.saveCache();
}

GCXMLEntry.prototype.getEntry = function()
{
	if (!this.entry && !this.loadCache()) return null;
	return this.entry;
}
GCXMLEntry.prototype.toRFC3339 = function(aDate)
{
	//if (!this.entry && !this.loadCache()) return null;
	var date = aDate ? aDate : new Date();
	var str = date.getUTCFullYear()                         
					+ "-" + ("0"+(date.getUTCMonth()+1)).substr(-2)
					+ "-" + ("0"+date.getUTCDate()).substr(-2)
					+ "T" + ("0"+date.getUTCHours()).substr(-2)
					+ ":"	+ ("0"+date.getUTCMinutes()).substr(-2)
					+ ":" + ("0"+date.getUTCSeconds()).substr(-2)
					+ "."	+ ("00"+date.getUTCMilliseconds()).substr(-3)
					+ "Z";
					
	return str;
}

GCXMLEntry.prototype.getId = function()
{
	if (!this.entry && !this.loadCache()) return "";
	var atom = this.atom;
	return (this.entry.atom::id).toString();
}

GCXMLEntry.prototype.getTitle = function()
{
	if (!this.entry && !this.loadCache()) return "";
	var atom = this.atom;
	return (this.entry.atom::title).toString();
}


GCXMLEntry.prototype.getEditLink = function()
{
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	return this.entry.atom::link.(@rel=="edit").@href;	
}

GCXMLEntry.prototype.getSelfLink = function()
{
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	return this.entry.atom::link.(@rel=="self").@href;
}

GCXMLEntry.prototype.getEntryEtag = function()
{
	if (!this.entry && !this.loadCache()) return null;
	var gd = this.gd;
	try {
		return this.entry.@gd::etag.toString();
	} catch(e) {
		return null;
	}
}

GCXMLEntry.prototype.getUpdatedTimeInMilliSec = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	var updated = this.entry.atom::updated.toString();
	var updatedArray = updated.split("T");
	var dateArray = updatedArray[0].split("-");
	var timeArray = updatedArray[1].split(":");

	var secArray = timeArray[2].split(".");
	timeArray[2] = secArray[0].replace(/Z$/,"");
	if (secArray[1]) {		
		timeArray.push(secArray[1].replace(/Z$/,""));
	}
	
	dateArray = dateArray.concat(timeArray);
	//remove prefix 0
	for (var i=0; i<dateArray.length; i++) {
		dateArray[i] = dateArray[i].replace(/^0+/,"");
	}
	
	var ms = Date.UTC(dateArray[0],   //Year
									dateArray[1]-1, //Month
									dateArray[2],   //Day
									dateArray[3],   //Hours
									dateArray[4],   //Minutes
									dateArray[5],   //Seconds
									dateArray[6]);  //Milliseconds
	return ms;
}

GCXMLEntry.prototype.setTBOnlyItemsToGContact = function(abItem)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	
	var gd = this.gd;
	//var atom = this.atom;
	try {
		delete this.entry.gd::extendedProperty;
	} catch(e) {
	}
	var key = "";
	for (key in this.itemObj) {
		var itemTB = this.itemObj[key];
		/*
		var itemXML = null;
		try {
			itemXML = this.entry.gd::extendedProperty.(@name=={key}).@value;
		} catch(e) {
		}
		if (!itemXML || !itemXML.toString()) {
			var xmlPiece = <gd:extendedProperty xmlns:gd={gd}></gd:extendedProperty>
			xmlPiece.@name = key;
			this.entry.gd::extendedProperty += xmlPiece;
		}
		*/
		/*
		try {
			var xmlPiece = <gd:extendedProperty xmlns:gd={gd}></gd:extendedProperty>
			xmlPiece.@name = key;
			this.entry.gd::extendedProperty += xmlPiece;
		} catch(e) {
		}
		*/
		var itemValueArray = new Array();
		for (var i=0; i<itemTB.length; i++) {
			if (itemTB[i].indexOf("-") == 0) continue; //obsolete item
			//var itemValue = abItem[itemTB[i]];
			var itemValue = abItem instanceof Components.interfaces.nsIAbCard
											? abItem.getProperty(itemTB[i], null)
											: abItem[itemTB[i]];
			if (itemValue == null || itemValue == undefined) itemValue = "";
			itemValue = itemValue.toString().replace(/\//g,'_gcslash_'); //escape
			itemValueArray.push(itemTB[i]+"="+itemValue);
		}
		if (itemValueArray.length > 0) {
			try {
				var xmlPiece = <gd:extendedProperty xmlns:gd={gd}></gd:extendedProperty>
				xmlPiece.@name = key;
				this.entry.gd::extendedProperty += xmlPiece;
				this.entry.gd::extendedProperty.(@name==key)[0].@value = itemValueArray.join("/");
			} catch(e) {
				dump(e+"\n");
			}
		}
	}
}

GCXMLEntry.prototype.loadTBOnlyItemsFromGContact = function(abItem)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	var gd = this.gd;
	var atom = this.atom;
	
	var key = "";
	for (key in this.itemObj) {
		var itemTB = this.itemObj[key];
		var itemXML = null;
		var itemValueStr = "";
		try {
			itemXML = this.entry.gd::extendedProperty.(@name==key).@value;
		} catch(e) {
		}
		if (itemXML && itemXML.toString()) {
			itemValueStr = itemXML;
			if (!itemValueStr) continue;
			var itemValueArray = itemValueStr.split("/");
			if (itemValueStr.indexOf("=") != -1) { //0.2.60 or above
				for (var i=0; i<itemValueArray.length; i++) {
					var item = itemValueArray[i].replace(/_gcslash_/g,'/'); //unescape _gcslash_
					var itemNameAndValue = item.split("=");
					var itemName = itemNameAndValue.shift();
					var itemValue = itemNameAndValue.join("=");
					itemName = this.convertTB2PropertyToTB3(itemName);
					try {
						if (abItem instanceof Components.interfaces.nsIAbCard) {
							abItem.setProperty(itemName,itemValue);
						} else {
							abItem[itemName] = itemValue;
						}
						//abItem[itemName] = itemValue;
					} catch(e) {
						dump(e+"\n");
					}
				}
			} else { // 0.2.50 or before
				for (var i=0; i<itemTB.length; i++) {
					var itemName = itemTB[i];
					if (itemName.indexOf("-") == 0) continue; //obsolete item
					try {
						itemName = this.convertTB2PropertyToTB3(itemName);
						if (itemValueArray[i]) {
							//abItem[itemTB[i]] = itemValueArray[i].replace(/_gcslash_/g,'/'); //unescape _gcslash_;
							if (abItem instanceof Components.interfaces.nsIAbCard) {
								abItem.setProperty(itemName, itemValueArray[i].replace(/_gcslash_/g,'/')); //unescape _gcslash_;
							} else {
								abItem[itemTB[i]] = itemValueArray[i].replace(/_gcslash_/g,'/'); //unescape _gcslash_;
							}
						} else {
							//abItem[itemTB[i]] = "";
							if (abItem instanceof Components.interfaces.nsIAbCard) {
								abItem.setProperty(itemName, "");
							} else {
								abItem[itemTB[i]] = "";
							}
						}
					} catch(e) {
						dump(e+"\n");
					}
				}
			}
		}
	}
}

GCXMLEntry.prototype.convertTB2PropertyToTB3 = function(propName)
{
	return propName;
}

//Group
function GCGroupEntry(entry)
{
	var superConstructor = GCXMLEntry;
	superConstructor.apply(this, arguments);
	this.mailList = null;
	this.dumpstr = "";
	this.systemDispName = null;
	this.suffix = "";
	
	if (entry) this.entry = entry;
	else this.createXMLEntry();
	
	this.itemObj = {
		listItem: ["description", "listNickName"]
	};
	
	var gContact = this.gContact;
	if (this.isSystemGroup()) {	
		var atom = this.atom;
		var name = (this.entry.atom::title).toString();
		var groupId = this.getSystemGroupId();
		//switch (name) {
		switch (groupId) {
			//case "System Group: My Contacts":
			case "Contacts":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_mycontacts");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			//case "System Group: Friends":
			case "Friends":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_friends");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			//case "System Group: Family":
			case "Family":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_family");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			//case "System Group: Coworkers":
			case "Coworkers":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_coworkers");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			default:
				dump("Unknow system group: "+name+"\n");
				this.systemDispName = name;
				break;
		}
	}
}

//Inherit
GCClassInheriter.inherit(GCGroupEntry, GCXMLEntry);

GCGroupEntry.prototype.createXMLEntry = function()
{
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	this.shallowDumpStr = "";
	this.entry = 
		<atom:entry xmlns:atom={atom} xmlns:gd={gd} xmlns:openSearch={openSearch} xmlns:gContact={gContact}/>;

  this.entry.atom::category.@scheme = "http://schemas.google.com/g/2005#kind";
  this.entry.atom::category.@term = "http://schemas.google.com/contact/2008#group";
}

GCGroupEntry.prototype.createMailList = function(addrBookDir)
{
  var mailList = Components.classes["@mozilla.org/addressbook/directoryproperty;1"].createInstance();
  mailList = mailList.QueryInterface(Components.interfaces.nsIAbDirectory);
  mailList.isMailList = true;
  this.updateMailListFromGGroup(mailList);
	//addrBookDir.addMailList(mailList);
  return mailList;
}

GCGroupEntry.prototype.updateMailListFromGGroup = function(list, card)
{
	if (!list) return;
	var listName = this.getTitle();
	//list.dirName = listName+this.suffix;
	list.dirName = listName;
	list.listNickName = "";  
	list.description = ""; 
	//if (card) card.setProperty("DisplayName", listName+this.suffix);
	if (card) card.setProperty("DisplayName", listName);
	
	this.loadTBOnlyItemsFromGContact(list);
	
	this.dumpstr = this.dump(list);
	this.shallowDumpStr = this.dump(list, true);
}

GCGroupEntry.prototype.updateGGroupFromMailList = function(list)
{
//	if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	var atom = this.atom;
	if (!this.isSystemGroup()) {
		var listName = list.dirName;
		if (this.suffix) {
			var re = new RegExp(this.suffix + "$");
			listName = listName.replace(re, "");
		}
		this.entry.atom::title = listName;
	}
	this.setTBOnlyItemsToGContact(list);
	this.dumpstr = this.dump(list);
	this.shallowDumpStr = this.dump(list, true);
}

GCGroupEntry.prototype.setMailList = function(list, account)
{
	if (!list) return;
	this.mailList = list;
	this.dumpstr = this.dump(list);
	this.shallowDumpStr = this.dump(list, true);
	var uri = null;
	try {
		uri = list.URI;
	} catch(e) {
		uri = null;
	}
	if (account && uri && this.getId()) {
		nsPreferences.setUnicharPref("gmcont."+account+"."+uri+".groupId", this.getId());
	}
}

GCGroupEntry.prototype.getMailList = function()
{
	return this.mailList;
}

GCGroupEntry.prototype.clearMailListChildren = function()
{
	var length = this.mailList.addressLists.length;
	for (var i = length - 1; i>=0; i--) {
		try {
	    this.mailList.addressLists.removeElementAt(i);
    } catch(e) {
    }
	}
}

GCGroupEntry.prototype.isChanged = function(list, woList)
{
	var newDumpStr = this.dump(list, woList);
	dump(this.dumpstr+"\n");
	dump(newDumpStr+"\n");
	
	return woList ? this.shallowDumpStr != newDumpStr : this.dumpstr != newDumpStr; 
}

GCGroupEntry.prototype.setSuffix = function(suffix)
{
	this.suffix = suffix;
}

GCGroupEntry.prototype.getTitle = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var title = "";
	if (this.isSystemGroup()) {
		title = this.systemDispName;
	} else {
		var atom = this.atom;
		title = (this.entry.atom::title).toString() + this.suffix;
	}
	return title;
}

GCGroupEntry.prototype.getDispName = function()
{
	return this.getTitle();
}

GCGroupEntry.prototype.getSystemDispName = function()
{
	return this.systemDispName;
}

GCGroupEntry.prototype.getSystemName = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return "";
	var atom = this.atom;
	var title = this.entry.atom::title;
	/*
	if (this.isSystemGroup()) {
		var atom = this.atom;
		title = this.entry.atom::title;
	} else {
		//title = this.getTitle();
		var atom = this.atom;
		title = this.entry.atom::title;
	}
	*/
	return title.toString();
}

GCGroupEntry.prototype.getOriginalName = function()
{
	return this.getSystemName();
}

GCGroupEntry.prototype.getSystemGroupId = function()
{
//	if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	
	try {
		var systemId = this.entry.gContact::systemGroup.@id;
		return systemId.toString();
	} catch(e) {
		return null;
	}
	
	return null;
}

GCGroupEntry.prototype.isSystemGroup = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return false;
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	
	try {
		var systemId = this.entry.gContact::systemGroup.@id;
		if (systemId.toString()) return true;
	} catch(e) {
		return false;
	}
	
	return false;
}

GCGroupEntry.prototype.dump = function(list, woList)
{
	var name = list.dirName ? list.dirName : "";
	if (name && this.suffix) {
		var re = new RegExp(this.suffix + "$");
		name = name.replace(re,"");
	}
	var desc = list.description ? list.description : "";
	var nickName = list.listNickName ? list.listNickName : "";
	var addrsStr = "";
	var addrsArray = new Array();
	if (!woList && list.addressLists) {
    //var total = list.addressLists.Count();
		var total = list.addressLists.length;
    if (total) {
      for ( var i = 0;  i < total; i++ ) {
        //var card = list.addressLists.GetElementAt(i);
      	var card = list.addressLists.queryElementAt(i, Components.interfaces.nsIAbCard)
        //card = card.QueryInterface(Components.interfaces.nsIAbCard);
        addrsArray.push(card.getProperty("PrimaryEmail", ""));
      }
    }
    addrsStr = addrsArray.sort().join("");
  }

  return name + desc + nickName + addrsStr;
}

//Contact
function GCContactEntry(entry)
{
	var superConstructor = GCXMLEntry;
	superConstructor.apply(this, arguments);
	this.card = null;
	this.dumpstr = "";
	this.photoManager = null;
	if (entry) this.entry = entry;
	else this.createXMLEntry();
	//update items only in TB
	this.itemObj = {
		contactTB: [
			//Contact
			"-FirstName",
			"-LastName",
			"-NickName",
			"-SecondEmail",
			"-aimScreenName", //aimScreenName is not a TB only item. Rename for compatibility
			"PreferMailFormat",
			//"AllowRemoteContent",
			"allowRemoteContent",
			"-PhoneticFirstName",
			"-PhoneticLastName"
		],
		homeAddrTB: [
			//Address - Home
			"HomeAddress2",
			"HomeCity",
			"HomeState",
			"HomeCountry",
			"HomeZipCode",
			"-WebPage2",
			"-Birthday",
			"-BirthMonth",
			"-BirthYear",
			"-Age"
		],
		workAddrTB: [
			//Address - Work
			"-Department",
			"-Company",
			"WorkAddress2",
			"WorkCity",
			"WorkState",
			"WorkZipCode",
			"WorkCountry",
			"-WebPage1"
		],
		otherTB: [
			//Other
			"-Custom1",
			"-Custom2",
			"-Custom3",
			"-Custom4"
		]
	};
}

//Inherit
GCClassInheriter.inherit(GCContactEntry, GCXMLEntry);

GCContactEntry.pseudoEmailsCount = 0;

GCContactEntry.prototype.createXMLEntry = function()
{
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	
	this.entry = 
		<atom:entry xmlns:atom={atom} xmlns:gd={gd} xmlns:openSearch={openSearch} xmlns:gContact={gContact}/>;

  this.entry.atom::category.@scheme = "http://schemas.google.com/g/2005#kind";
  this.entry.atom::category.@term = "http://schemas.google.com/contact/2008#contact";
}

GCContactEntry.prototype.getABCard = function()
{
	return this.card;
}

GCContactEntry.prototype.setABCard = function(card)
{
	this.card = card;
	//this.dumpstr = this.card.convertToEscapedVCard();
	this.dumpstr = this.dump(this.card);
}

GCContactEntry.prototype.createABCard = function(addrBookDir)
{
	var card = Components.classes["@mozilla.org/addressbook/cardproperty;1"].
								createInstance(Components.interfaces.nsIAbCard);
	this.updateCardFromGContact(card);
	var abCard = addrBookDir.addCard(card);
	return abCard;
}

GCContactEntry.prototype.belongToGroup = function(gcGroup)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	var id = gcGroup.getId();
	var gContact = this.gContact;
	var group = null;
	try {
		group = this.entry.gContact::groupMembershipInfo.(@href==id);
	} catch(e) {
	}
	if (group.toString() != "") return;
	var xmlGrp = <gContact:groupMembershipInfo xmlns:gContact={gContact}></gContact:groupMembershipInfo>
	xmlGrp.@href = id;
	this.entry.gContact::groupMembershipInfo += xmlGrp;
	dump("Belong to group\n");
	//dump(this.entry.gContact::groupMembershipInfo+"\n");
}

GCContactEntry.prototype.unbelongToGroup = function(gcGroup)
{
	//if (!this.entry) this.loadCache();
		if (!this.entry && !this.loadCache()) return;
	var id = gcGroup.getId();
	var gContact = this.gContact;
	var group = null;
	try {
		group = this.entry.gContact::groupMembershipInfo.(@href==id)[0];
	} catch(e) {
	}
	if (group != null) delete this.entry.gContact::groupMembershipInfo.(@href==id)[0];
}

GCContactEntry.prototype.getBelongingGroupIds = function()
{
	//if (!this.entry) this.loadCache();
	var ids = new Array();
	if (!this.entry && !this.loadCache()) return ids;
	var gContact = this.gContact;
	var groups = this.entry.gContact::groupMembershipInfo;
	for (var i=0; i<groups.length(); i++) {
		var group = groups[i];
		try {
			ids.push(group.@href);
		} catch(e) {
			dump("No href\n");
		}
	}
	
	return ids;
}

GCContactEntry.prototype.setPhotoManager = function(manager)
{
	this.photoManager = manager;
}

GCContactEntry.prototype.setPhoto = function(photoURI)
{
  this.card.setProperty("PhotoType", "file");
  // Store the original URI
  this.card.setProperty("PhotoURI", photoURI);
  // Save the new file and store its URI as PhotoName 
  var file = this.photoManager.savePhoto(photoURI);
  if (file) {
		// Remove the original, if any
    this.photoManager.removePhoto(this.card.getProperty("PhotoName", null));
    this.card.setProperty("PhotoName", file.leafName);
  } else {
  	dump("save photo failed\n");
  }
	return this.card;
}

GCContactEntry.prototype.resetPhoto = function()
{
  try {
  	this.card.setProperty("PhotoType", "generic");
  	this.card.setProperty("PhotoURI", "");
  	this.photoManager.removePhoto(this.card.getProperty("PhotoName", null));
  	this.card.setProperty("PhotoName", "");
  }catch(e){
  	dump(e);
  }
  
  return this.card;
}

GCContactEntry.prototype.updateCardFromGContact = function(abCard)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return false;
	var gd = this.gd;
	var gContact = this.gContact;
	var atom = this.atom;
	abCard.QueryInterface(Components.interfaces.nsIAbCard);

	var oldEtag = abCard.getProperty("GmContactEtag",null);
	var newEtag = this.getEntryEtag();
	if (oldEtag && newEtag && oldEtag == newEtag) {
		this.dumpstr = this.dump(abCard);
		return;
	} else if (newEtag) {
		abCard.setProperty("GmContactEtag", newEtag);
	}

	//Contact tab
	//display name
	/*
	try {
		this.entry.atom::title.@type = "text";
		//abCard.displayName = this.entry.atom::title;
		abCard.setProperty("DisplayName",
											(this.entry.atom::title).toString()
											);
	} catch(e) {
		//dump("no disp name\n");		
	}
	*/
	//Name
	try {
		var fullName = (this.entry.gd::name.gd::fullName).toString();
		//var fullName = (this.entry.atom::title).toString();
		abCard.setProperty("DisplayName", fullName);
	} catch(e) {
		abCard.setProperty("DisplayName", "");
	}
	
	//var deriveNameParts = nsPreferences.getBoolPref("gmcont.use_structured_name", false);
	try {
		//var givenName = deriveNameParts ?
		//								(this.entry.gd::name.gd::givenName).toString() :
		//								(this.entry.gd::givenName).toString();
		var givenName = (this.entry.gd::name.gd::givenName).toString();
		abCard.setProperty("FirstName", givenName);
	} catch(e) {
		abCard.setProperty("FirstName", "");
	}
	
	try {
		//var familyName = deriveNameParts ?
		//								(this.entry.gd::name.gd::familyName).toString() :
		//								(this.entry.gd::familyName).toString();
		var familyName = (this.entry.gd::name.gd::familyName).toString();
		abCard.setProperty("LastName", familyName);
	} catch(e) {
		abCard.setProperty("LastName", "");
	}

	try {
		var givenNameYomi = (this.entry.gd::name.gd::givenName.@yomi).toString();
		abCard.setProperty("PhoneticFirstName", givenNameYomi);
	} catch(e) {
		abCard.setProperty("PhoneticFirstName", "");
	}
	
	try {
		var familyNameYomi = (this.entry.gd::name.gd::familyName.@yomi).toString();
		abCard.setProperty("PhoneticLastName", familyNameYomi);
	} catch(e) {
		abCard.setProperty("PhoneticLastName", "");
	}

	try {
		var nickName = (this.entry.gContact::nickname).toString();
		abCard.setProperty("NickName", nickName);
	} catch(e) {
		abCard.setProperty("NickName", "");
	}
	
	//email
	//primary email
	try {
		//if (this.entry.gd::email.(@primary=="true").length() > 0) {
		if (this.entry.gd::email.(function::attribute("primary")=="true").length() > 0) {
			abCard.setProperty("PrimaryEmail",
												//(this.entry.gd::email.(@primary=="true").@address).toString()
												(this.entry.gd::email.(function::attribute("primary")=="true").@address).toString()
												);
		} else if (this.entry.gd::email.@address.length() > 0) {
			abCard.setProperty("PrimaryEmail",
												(this.entry.gd::email.@address[0]).toString()
												);
		} else {
			abCard.setProperty("PrimaryEmail", "");
		}
	} catch(e) {
		abCard.setProperty("PrimaryEmail", "");
	}
	
	var addPseudoEmail = nsPreferences.getBoolPref("gmcont.add_pseudo_email", false);
	if (addPseudoEmail && !abCard.getProperty("PrimaryEmail", null)) {
		GCContactEntry.pseudoEmailsCount++;
		abCard.setProperty("PrimaryEmail",
												GCContactEntry.pseudoEmailsCount+"@pseudo.googlecontacts");
	}
	
	//second email
	try {
		//abCard.secondEmail = this.entry.gd::email[1].@address;
		abCard.setProperty("SecondEmail",
											(this.entry.gd::email[1].@address).toString()
											);
	} catch(e) {
		abCard.setProperty("SecondEmail", "");
		//dump("no second email\n");
	}
	
	try {
		//abCard.aimScreenName = this.entry.gd::im[0].@address;
		abCard.setProperty("_AimScreenName",
											(this.entry.gd::im[0].@address).toString()
											);
	} catch(e) {
		abCard.setProperty("_AimScreenName", "");
	}
	
	//phone number
	try {
		//abCard.workPhone = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#work")[0];
		abCard.setProperty("WorkPhone",
											(this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#work")[0]).toString()
											);
	} catch(e) {
		abCard.setProperty("WorkPhone", "");
	}
	
	try {
		 //abCard.homePhone = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#home")[0];
		abCard.setProperty("HomePhone",
											(this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#home")[0]).toString()
											);
	} catch(e) {
		abCard.setProperty("HomePhone", "");
	}

	try {
		 //abCard.faxNumber = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#home_fax")[0];
		abCard.setProperty("FaxNumber",
											(this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#home_fax")[0]).toString()
											);
	} catch(e) {
		abCard.setProperty("FaxNumber", "");
	}

	try {
		 //abCard.pagerNumber = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#pager")[0];
		abCard.setProperty("PagerNumber",
											(this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#pager")[0]).toString()
											);
	} catch(e) {
		abCard.setProperty("PagerNumber", "");
	}
	
	try {
		 //abCard.cellularNumber = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#mobile")[0];
		abCard.setProperty("CellularNumber",
											(this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#mobile")[0]).toString()
											);
	} catch(e) {
		abCard.setProperty("CellularNumber", "");
	}
	
	this.loadTBOnlyItemsFromGContact(abCard);
	
	//Address tab
	//home
	var addrFields = [];
	try {
		//var homeAddr = (this.entry.gd::postalAddress.(@rel=="http://schemas.google.com/g/2005#home")[0]).toString();
		var homeAddr = (this.entry.gd::structuredPostalAddress.(@rel=="http://schemas.google.com/g/2005#home")[0]).gd::formattedAddress.toString();
		addrFields = this.parseGContactAddr(homeAddr, "Home");
	} catch(e) {
		addrFields = this.parseGContactAddr(null, "Home");
	}
	for (var i=0; i<addrFields.length; i++) {
		var field = addrFields[i];
		if (field) abCard.setProperty(field.prop, field.value);
	}


	//work
	addrFields = [];
	try {
		//var workAddr = (this.entry.gd::postalAddress.(@rel=="http://schemas.google.com/g/2005#work")[0]).toString();
		var workAddr = (this.entry.gd::structuredPostalAddress.(@rel=="http://schemas.google.com/g/2005#work")[0]).gd::formattedAddress.toString();
		addrFields = this.parseGContactAddr(workAddr, "Work");
	} catch(e) {
		addrFields = this.parseGContactAddr(null, "Work");
	}
	for (var i=0; i<addrFields.length; i++) {
		var field = addrFields[i];
		if (field) abCard.setProperty(field.prop, field.value);
	}
	
	var jobTitle = "";
	try {
		if (this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other").length() > 0) {
			jobTitle = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgTitle;
		} else if (this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work").length() > 0) {
			jobTitle = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work")[0].gd::orgTitle;
		}
		abCard.setProperty("JobTitle", jobTitle.toString());
	} catch(e) {
		abCard.setProperty("JobTitle", "");
	}

	var company = "";
	try {
		if (this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other").length() > 0) {
			company = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgName;
		} else if (this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work").length() > 0) {
			company = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work")[0].gd::orgName;
		}
		abCard.setProperty("Company", company.toString());
	} catch(e) {
		abCard.setProperty("Company", "");
	}
	
	var department = "";
	try {
		if (this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other").length() > 0) {
			department = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgDepartment;
		} else if (this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work").length() > 0) {
			department = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#work")[0].gd::orgDepartment;
		}
		abCard.setProperty("Department", department.toString());
	} catch(e) {
		abCard.setProperty("Department", "");
	}
	
	//Birthday
	try {
			var birthday = (this.entry.gContact::birthday.@when).toString();
			if (birthday) {
				//birthday = YYYY-MM-DD or --MM-DD
				var birthdayFields = birthday.split("-");
				abCard.setProperty("BirthDay", parseInt(birthdayFields.pop(), 10));
				abCard.setProperty("BirthMonth", parseInt(birthdayFields.pop(), 10));
				var year = birthdayFields.pop();
				if (year) abCard.setProperty("BirthYear", parseInt(year, 10));
				//abCard.setProperty("BirthYear", parseInt(birthdayFields[0], 10));
				//abCard.setProperty("BirthMonth", parseInt(birthdayFields[1], 10));
				//abCard.setProperty("BirthDay", parseInt(birthdayFields[2], 10));
			} else {
				abCard.setProperty("BirthYear", "");
				abCard.setProperty("BirthMonth", "");
				abCard.setProperty("BirthDay", "");
			}
	} catch(e) {
		abCard.setProperty("BirthYear", "");
		abCard.setProperty("BirthMonth", "");
		abCard.setProperty("BirthDay", "");
	}

	//web
	//work
	try {			
			if (this.entry.gContact::website.(@rel=="work").length() > 0) {
				var web1 = this.entry.gContact::website.(@rel=="work")[0];
				abCard.setProperty("WebPage1", (web1.@href).toString());
			} else {
				abCard.setProperty("WebPage1", "");
			}
	} catch(e) {
		abCard.setProperty("WebPage1", "");
	}
	
	try {
			if (this.entry.gContact::website.(@rel=="home").length() > 0) {
				var web2 = this.entry.gContact::website.(@rel=="home")[0];
				abCard.setProperty("WebPage2", (web2.@href).toString());
			} else {
				abCard.setProperty("WebPage2", "");
			}
	} catch(e) {
		abCard.setProperty("WebPage2", "");
	}
	
	//Other tab
	//custom1
	try {
			if (this.entry.gContact::userDefinedField.(@key=="Custom1").length() > 0) {
				var cust1 = this.entry.gContact::userDefinedField.(@key=="Custom1")[0];
				abCard.setProperty("Custom1", (cust1.@value).toString());
			} else {
				abCard.setProperty("Custom1", "");
			}
	} catch(e) {
		abCard.setProperty("Custom1", "");
	}
	
	//custom2
	try {
			if (this.entry.gContact::userDefinedField.(@key=="Custom2").length() > 0) {
				var cust2 = this.entry.gContact::userDefinedField.(@key=="Custom2")[0];
				abCard.setProperty("Custom2", (cust2.@value).toString());
			} else {
				abCard.setProperty("Custom2", "");
			}
	} catch(e) {
		abCard.setProperty("Custom2", "");
	}
	
	//custom3
	try {
			if (this.entry.gContact::userDefinedField.(@key=="Custom3").length() > 0) {
				var cust3 = this.entry.gContact::userDefinedField.(@key=="Custom3")[0];
				abCard.setProperty("Custom3", (cust3.@value).toString());
			} else {
				abCard.setProperty("Custom3", "");
			}
	} catch(e) {
		abCard.setProperty("Custom3", "");
	}
	
	//custom4
	try {
			if (this.entry.gContact::userDefinedField.(@key=="Custom4").length() > 0) {
				var cust4 = this.entry.gContact::userDefinedField.(@key=="Custom4")[0];
				abCard.setProperty("Custom4", (cust4.@value).toString());
			} else {
				abCard.setProperty("Custom4", "");
			}
	} catch(e) {
		abCard.setProperty("Custom4", "");
	}
	
	//notes
	try {
		//abCard.notes = this.entry.atom::content;
		abCard.setProperty("Notes",
											(this.entry.atom::content).toString()
											);
	} catch(e) {
		abCard.setProperty("Notes", "");
	}
	
	//this.dumpstr = abCard.convertToEscapedVCard();
	this.dumpstr = this.dump(abCard);

}

GCContactEntry.prototype.parseGContactAddr = function(addr, type) 
{
	var fields = new Array();
	var format = nsPreferences.copyUnicharPref("gmcont.addr_conv_format", "");
	var delimiter = nsPreferences.copyUnicharPref("gmcont.addr_conv_delimiter", ", ");
	if (delimiter == '<BR>') delimiter = "\n";
	
	if (format && type) {
		var items = format.split(",");
		var addrFields = addr ? addr.split(delimiter) : [];
		for (var i=0; i<items.length; i++) {
			var val = addrFields[i] ? addrFields[i] : "";
			var field = {
				prop: type+items[i],
				value: val
			};
			fields.push(field);
		}
	} else if (!format && type) { //all in Address
			var field = {
				prop: type+"Address",
				value: addr
			};
			fields.push(field);
	}
	
	return fields;
}

GCContactEntry.prototype.updateGContactFromCard = function(abCard)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	var atom = this.atom;
	var gd = this.gd;
	var gContact = this.gContact;

	abCard.QueryInterface(Components.interfaces.nsIAbCard);
	if (!abCard.getProperty("DisplayName", null) && !abCard.getProperty("PrimaryEmail", null)) {
		var msg = this.localeBundle.getString("gmcont.no_primary_email");
		alert(msg);
		return false;
	}
	
	//Contact tab
	//display name
	/*
	var dispName = abCard.getProperty("DisplayName", null); 
	if (dispName) {
		this.entry.atom::title = "";
		this.entry.atom::title = dispName;
		this.entry.atom::title.@type = "text";
	} else {
		try {
			delete this.entry.atom::title;			
		} catch(e) {
		}
	}
	*/
	//Name
	var fullName = abCard.getProperty("DisplayName", null); 
	if (fullName) {
		this.entry.gd::name = "";
		this.entry.gd::name.gd::fullName = fullName;
	} else {
		try {
			delete this.entry.gd::name.gd::fullName;
		} catch(e) {
		}
	}
	
	var givenName = abCard.getProperty("FirstName", null); 
	if (givenName) {
		//this.entry.gd::name = "";
		this.entry.gd::name.gd::givenName = givenName;
		//this.entry.gd::givenName = givenName;
	} else {
		try {
			delete this.entry.gd::name.gd::givenName;
			//delete this.entry.gd::givenName;			
		} catch(e) {
		}
	}
	
	var familyName = abCard.getProperty("LastName", null); 
	if (familyName) {
		//this.entry.gd::name = "";
		this.entry.gd::name.gd::familyName = familyName;
		//this.entry.gd::familyName = familyName;
	} else {
		try {
			delete this.entry.gd::name.gd::faimlyName;
			//delete this.entry.gd::faimlyName;			
		} catch(e) {
		}
	}

	var givenNameYomi = abCard.getProperty("PhoneticFirstName", null); 
	if (givenNameYomi && (this.entry.gd::name.gd::givenName).toString()) {
		this.entry.gd::name.gd::givenName.@yomi = givenNameYomi;
	} else {
		try {
			//delete this.entry.gd::name.gd::givenName;			
			delete this.entry.gd::name.gd::givenName.@yomi;			
		} catch(e) {
		}
	}

	var familyNameYomi = abCard.getProperty("PhoneticLastName", null); 
	if (familyNameYomi && (this.entry.gd::name.gd::familyName).toString()) {
		this.entry.gd::name.gd::familyName.@yomi = familyNameYomi;
	} else {
		try {
			//delete this.entry.gd::name.gd::familyName;			
			delete this.entry.gd::name.gd::familyName.@yomi;			
		} catch(e) {
		}
	}
	
	var nickName = abCard.getProperty("NickName", null); 
	if (nickName) {
		this.entry.gContact::nickname = nickName;
	} else {
		try {
			delete this.entry.gContact::nickname;
		} catch(e) {
		}
	}
	
	//email
	var delPrimary = false;
	var pAddr = abCard.getProperty("PrimaryEmail", null);
	if (pAddr && pAddr.indexOf("@pseudo.googlecontacts") != -1) pAddr = "pseudo";
	if (pAddr && pAddr != "pseudo") {
		this.entry.gd::email[0] = "";
		this.entry.gd::email[0].@address = pAddr;
		this.entry.gd::email[0].@primary = "true";
		try {
			var rel = this.entry.gd::email[0].@rel;
			if (!rel.toString()) this.entry.gd::email[0].@rel = "http://schemas.google.com/g/2005#other";
		} catch(e) {
			this.entry.gd::email[0].@rel = "http://schemas.google.com/g/2005#other";
		}
	} else {
		delPrimary = true;
		/*
		try {
			delete this.entry.gd::email[0];
		} catch(e) {
		}
		*/	
	}
	if (!pAddr){
		var addPseudoEmail = nsPreferences.getBoolPref("gmcont.add_pseudo_email", false);
		if (addPseudoEmail) {
			GCContactEntry.pseudoEmailsCount++;
			abCard.setProperty("PrimaryEmail",
													GCContactEntry.pseudoEmailsCount+"@pseudo.googlecontacts");
		}
	}

	var delSecondary = false;
	var sAddr = abCard.getProperty("SecondEmail", null);
	if (sAddr) {
		try {
			this.entry.gd::email[1].@address = sAddr;
		} catch(e) {
			var xmlEmail2 = <gd:email xmlns:gd={gd}/>;
			xmlEmail2.@address = sAddr;
			xmlEmail2.@rel = "http://schemas.google.com/g/2005#other";			
			this.entry.gd::email += xmlEmail2;
		}
	} else {
		delSecondary = true;
		/*
		try {
			delete this.entry.gd::email[1];
		} catch(e) {
		}
		*/
	}

	if (delSecondary) {
		try {
			delete this.entry.gd::email[1];
		} catch(e) {
		}
	}
	if (delPrimary) {
		try {
			delete this.entry.gd::email[0];
			//secondary email should be primary
			if (sAddr) {
				dump("No primary email. Secondary one becomes a primary.\n")
				this.entry.gd::email[0].@primary = "true";
				abCard.setProperty("PrimaryEmail", sAddr);
				//set new secondary
				dump("Set new secondary email to a card.\n")
				if (this.entry.gd::email.length() > 1) {
					abCard.setProperty("SecondEmail", this.entry.gd::email[1].@address.toString());
				} else {
					abCard.setProperty("SecondEmail", "");
				}
			}
		} catch(e) {
		}
		
	}
	
	var aim = abCard.getProperty("_AimScreenName", null); 
	if (aim) {
		try {
			this.entry.gd::im[0].@address = aim;
		} catch(e) {
			var xmlIM = <gd:im xmlns:gd={gd}/>;
			xmlIM.@address = aim;
			xmlIM.@protocol='http://schemas.google.com/g/2005#GOOGLE_TALK';
			xmlIM.@rel = "http://schemas.google.com/g/2005#other";			
			this.entry.gd::im += xmlIM;
		}
	} else {
		try {
			delete this.entry.gd::im[0];
		} catch(e) {
		}
	}
	
	//phone number
	var phoneNum = null;
	try {
		phoneNum = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#work");
	} catch(e) {
	}
	
	var wPhone = abCard.getProperty("WorkPhone", null); 
	if (wPhone) {
		if (phoneNum && phoneNum.toString()) {
			phoneNum[0] = wPhone;
		} else { //add new phone number
			var xmlPhone = <gd:phoneNumber xmlns:gd={gd}>{wPhone}</gd:phoneNumber>
			xmlPhone.@rel = "http://schemas.google.com/g/2005#work";
			this.entry.gd::phoneNumber += xmlPhone;
		}
	} else {//delete existing phone number
		try {
			if (phoneNum && phoneNum.toString()) delete phoneNum[0];
		} catch(e) {
		}
	}
	
	phoneNum = null;
	try {
		phoneNum = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#home");
	} catch(e) {
	}
	
	var hPhone = abCard.getProperty("HomePhone", null); 
	if (hPhone) {
		if (phoneNum && phoneNum.toString()) {
			phoneNum[0] = hPhone;
		} else { //add new phone number
			var xmlPhone = <gd:phoneNumber xmlns:gd={gd}>{hPhone}</gd:phoneNumber>
			xmlPhone.@rel = "http://schemas.google.com/g/2005#home";
			this.entry.gd::phoneNumber += xmlPhone;
		}
	} else {//delete existing phone number
		try {
			if (phoneNum && phoneNum.toString()) delete phoneNum[0];
		} catch(e) {
		}
	}
	
	phoneNum = null;
	try {
		phoneNum = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#home_fax");
	} catch(e) {
	}
	
	var fNum = abCard.getProperty("FaxNumber", null); 
	if (fNum) {
		if (phoneNum && phoneNum.toString()) {
			phoneNum[0] = fNum;
		} else { //add new phone number
			var xmlPhone = <gd:phoneNumber xmlns:gd={gd}>{fNum}</gd:phoneNumber>
			xmlPhone.@rel = "http://schemas.google.com/g/2005#home_fax";
			this.entry.gd::phoneNumber += xmlPhone;
		}
	} else {//delete existing phone number
		try {
			if (phoneNum && phoneNum.toString()) delete phoneNum[0];
		} catch(e) {
		}
	}

	phoneNum = null;
	try {
		phoneNum = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#pager");
	} catch(e) {
	}
	
	var pgNum = abCard.getProperty("PagerNumber", null); 
	if (pgNum) {
		if (phoneNum && phoneNum.toString()) {
			phoneNum[0] = pgNum;
		} else { //add new phone number
			var xmlPhone = <gd:phoneNumber xmlns:gd={gd}>{pgNum}</gd:phoneNumber>
			xmlPhone.@rel = "http://schemas.google.com/g/2005#pager";
			this.entry.gd::phoneNumber += xmlPhone;
		}
	} else {//delete existing phone number
		try {
			if (phoneNum && phoneNum.toString()) delete phoneNum[0];
		} catch(e) {
		}
	}

	phoneNum = null;
	try {
		phoneNum = this.entry.gd::phoneNumber.(@rel=="http://schemas.google.com/g/2005#mobile");
	} catch(e) {
	}
	
	var cellNum = abCard.getProperty("CellularNumber", null);
	if (cellNum) {
		if (phoneNum && phoneNum.toString()) {
			phoneNum[0] = cellNum;
		} else { //add new phone number
			var xmlPhone = <gd:phoneNumber xmlns:gd={gd}>{cellNum}</gd:phoneNumber>
			xmlPhone.@rel = "http://schemas.google.com/g/2005#mobile";
			this.entry.gd::phoneNumber += xmlPhone;
		}
	} else {//delete existing phone number
		try {
			if (phoneNum && phoneNum.toString()) delete phoneNum[0];
		} catch(e) {
		}
	}

	//Address tab
	//home
	//convert address fields
	var addrStr = this.parseTBAddr(abCard, "Home");
	var addrText = null;
	try {
		//addrText = this.entry.gd::postalAddress.(@rel == "http://schemas.google.com/g/2005#home");
		addrText = this.entry.gd::structuredPostalAddress.(@rel == "http://schemas.google.com/g/2005#home");
	} catch(e) {
	}
	var addrValue = addrStr ? addrStr : abCard.getProperty("HomeAddress", null);
	if (addrValue) {
		if (addrText && addrText.toString()) {
			//addrText[0] = addrValue;
			addrText[0].gd::formattedAddress = addrValue;
		} else {
			//var xmlAddr = <gd:postalAddress xmlns:gd={gd}>{addrValue}</gd:postalAddress>
			//xmlAddr.@rel = "http://schemas.google.com/g/2005#home";
			//this.entry.gd::postalAddress += xmlAddr;
			var xmlAddr = <gd:structuredPostalAddress xmlns:gd={gd} rel={"http://schemas.google.com/g/2005#home"}><gd:formattedAddress>{addrValue}</gd:formattedAddress></gd:structuredPostalAddress>
			this.entry.gd::structuredPostalAddress += xmlAddr;
		}
	} else {
		try {
			if (addrText && addrText.toString()) delete addrText[0];
		} catch(e) {
		}
	}

	//work
	var addrStr = this.parseTBAddr(abCard, "Work");
	addrText = null;
	try {
		//addrText = this.entry.gd::postalAddress.(@rel == "http://schemas.google.com/g/2005#work");
		addrText = this.entry.gd::structuredPostalAddress.(@rel == "http://schemas.google.com/g/2005#work");
	} catch(e) {
	}
	var addrValue = addrStr ? addrStr : abCard.getProperty("WorkAddress", null);
	if (addrValue) {
		if (addrText && addrText.toString()) {
			//addrText[0] = addrValue;
			addrText[0].gd::formattedAddress = addrValue;
		} else {
			//var xmlAddr = <gd:postalAddress xmlns:gd={gd}>{addrValue}</gd:postalAddress>
			//xmlAddr.@rel = "http://schemas.google.com/g/2005#work";
			//this.entry.gd::postalAddress += xmlAddr;
			var xmlAddr = <gd:structuredPostalAddress xmlns:gd={gd} rel={"http://schemas.google.com/g/2005#work"}><gd:formattedAddress>{addrValue}</gd:formattedAddress></gd:structuredPostalAddress>
			this.entry.gd::structuredPostalAddress += xmlAddr;
		}
	} else {
		try {
			if (addrText && addrText.toString()) delete addrText[0];
		} catch(e) {
		}
	}

	var jTitle = abCard.getProperty("JobTitle", null); 
	if (jTitle) {
		try {
			this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgTitle = jobTitle;
		} catch(e) {
			this.entry.gd::organization = "";
			this.entry.gd::organization.@rel = "http://schemas.google.com/g/2005#other";
			this.entry.gd::organization.gd::orgTitle = jTitle;
		}
	} else {
		try {
			delete this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgTitle;
		} catch(e) {
		}
	}
	
	var cmp = abCard.getProperty("Company", null);  
	if (cmp) {
		try {
			this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgName = cmp;
		} catch(e) {
			this.entry.gd::organization = "";
			this.entry.gd::organization.@rel = "http://schemas.google.com/g/2005#other";
			this.entry.gd::organization.gd::orgName = cmp;
		}
	} else {
		try {
			delete this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgName;
		} catch(e) {
		}		
	}

	var department = abCard.getProperty("Department", null);  
	if (department) {
		try {
			this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgDepartment = department;
		} catch(e) {
			this.entry.gd::organization = "";
			this.entry.gd::organization.@rel = "http://schemas.google.com/g/2005#other";
			this.entry.gd::organization.gd::orgDepartment = department;
		}
	} else {
		try {
			delete this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].gd::orgDepartment;
		} catch(e) {
		}		
	}
	
	try {
		var org = this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0].*;
		if (org.length() == 0) delete this.entry.gd::organization.(@rel=="http://schemas.google.com/g/2005#other")[0];
	} catch(e) {
	}
	
	//Birthday
	var year = abCard.getProperty("BirthYear", null);
	if (!year) year = "-";
	var month = abCard.getProperty("BirthMonth", null);
	var day = abCard.getProperty("BirthDay", null);
	if (month && day) {
		try {
			this.entry.gContact::birthday.@when = year+"-"+month+"-"+day;
		} catch(e){
			this.entry.gContact::birthday="";
			this.entry.gContact::birthday.@when = year+"-"+month+"-"+day;
		}
	} else {
		try {
			delete this.entry.gContact::birthday;
		} catch(e){
		}
	}

	//web
	//work
	try {
		var webVal1 = abCard.getProperty("WebPage1", null); 
		var web1 = this.entry.gContact::website.(@rel == "work")[0];
		if (webVal1) {
			if (web1 != undefined) {
				web1.@href = webVal1;
			} else {
				var xmlWeb1 = <gContact:website xmlns:gContact={gContact} rel={"work"} href={webVal1}></gContact:website>
				this.entry.gContact::website += xmlWeb1;				
			}
		} else {
			try {
				if (web1 != undefined) delete this.entry.gContact::website.(@rel == "work")[0];
			} catch(e) {
			}
		}
	} catch(e) {
	}
	
	//home
	try {
		var webVal2 = abCard.getProperty("WebPage2", null); 
		var web2 = this.entry.gContact::website.(@rel == "home")[0];
		if (webVal2) {
			if (web2 != undefined) {
				web2.@href = webVal2
			} else {
				var xmlWeb2 = <gContact:website xmlns:gContact={gContact} rel={"home"} href={webVal2}></gContact:website>
				this.entry.gContact::website += xmlWeb2;				
			}
		} else {
			try {
				if (web2 != undefined) delete this.entry.gContact::website.(@rel == "home")[0];
			} catch(e) {
			}
		}
	} catch(e) {
	}
	
	//Other tab
	//custom1
	try {
		var custVal1 = abCard.getProperty("Custom1", null);
		var cust1 = this.entry.gContact::userDefinedField.(@key == "Custom1")[0];
		if (custVal1) {
			if (cust1 != undefined) {
				cust1.@value = custVal1;
			} else {
				var xmlCust1 = <gContact:userDefinedField xmlns:gContact={gContact} key={"Custom1"} value={custVal1}></gContact:userDefinedField>
				this.entry.gContact::userDefinedField += xmlCust1;				
			}
		} else {
			try {
				if (cust1 != undefined) delete this.entry.gContact::userDefinedField.(@key == "Custom1")[0];			
			} catch(e) {
				dump(e);
			}
		}
	} catch(e) {
	}
	
	//custom2
	try {
		var custVal2 = abCard.getProperty("Custom2", null);
		var cust2 = this.entry.gContact::userDefinedField.(@key == "Custom2")[0];
		if (custVal2) {
			if (cust2 != undefined) {
				cust2.@value = custVal2;
			} else {
				var xmlCust2 = <gContact:userDefinedField xmlns:gContact={gContact} key={"Custom2"} value={custVal2}></gContact:userDefinedField>
				this.entry.gContact::userDefinedField += xmlCust2;				
			}
		} else {
			try {
				if (cust2 != undefined) delete this.entry.gContact::userDefinedField.(@key == "Custom2")[0];			
			} catch(e) {
				dump(e);
			}
		}
	} catch(e) {
	}
	
	//custom3
	try {
		var custVal3 = abCard.getProperty("Custom3", null);
		var cust3 = this.entry.gContact::userDefinedField.(@key == "Custom3")[0];
		if (custVal3) {
			if (cust3 != undefined) {
				cust3.@value = custVal3;
			} else {
				var xmlCust3 = <gContact:userDefinedField xmlns:gContact={gContact} key={"Custom3"} value={custVal3}></gContact:userDefinedField>
				this.entry.gContact::userDefinedField += xmlCust3;				
			}
		} else {
			try {
				if (cust3 != undefined) delete this.entry.gContact::userDefinedField.(@key == "Custom3")[0];			
			} catch(e) {
			}
		}
	} catch(e) {
		dump(e);
	}
	
	//custom4
	try {
		var custVal4 = abCard.getProperty("Custom4", null);
		var cust4 = this.entry.gContact::userDefinedField.(@key == "Custom4")[0];
		if (custVal4) {
			if (cust4 != undefined) {
				cust4.@value = custVal4;
			} else {
				var xmlCust4 = <gContact:userDefinedField xmlns:gContact={gContact} key={"Custom4"} value={custVal4}></gContact:userDefinedField>
				this.entry.gContact::userDefinedField += xmlCust4;				
			}
		} else {
			try {
				if (cust4 != undefined) delete this.entry.gContact::userDefinedField.(@key == "Custom4")[0];			
			} catch(e) {
			}
		}
	} catch(e) {
	}
	
	//notes
	try {
		var nts = abCard.getProperty("Notes", null); 
		if (nts) this.entry.atom::content = nts;
		else delete this.entry.atom::content;
	} catch(e) {
	}
	
	//TB only items
	this.setTBOnlyItemsToGContact(abCard);
	
	//this.dumpstr = abCard.convertToEscapedVCard();
	this.dumpstr = this.dump(abCard);

	//set update date
	/*
	try {
		var oldDate = this.entry.atom::updated;
		dump(oldDate+"\n");
		var newDate = this.toRFC3339();
		if (oldDate) this.entry.atom::updated = newDate;
		dump(newDate+"\n");
	} catch(e) {
		dump(e+"\n");
	}
*/
	return true;
}

GCContactEntry.prototype.parseTBAddr = function(abCard, type) 
{
	var fields = new Array();
	var format = nsPreferences.copyUnicharPref("gmcont.addr_conv_format", "");
	var delimiter = nsPreferences.copyUnicharPref("gmcont.addr_conv_delimiter", ", ");
	if (delimiter == '<BR>') delimiter = "\n";
	var isEmpty = true;
	if (format) {
		var items = format.split(",");
		for (var i=0; i<items.length; i++) {
			var value = abCard.getProperty(type+items[i], null);
			if (!value) value = "";
			else isEmpty = false;
			//if (value) fields.push(value);
			fields.push(value);
		}
	}
	return isEmpty ? null : fields.join(delimiter);
}

GCContactEntry.prototype.convertTB2PropertyToTB3 = function(propName)
{
	var cardMaps = [];
	// Contact > Name
  cardMaps["firstName"] = "FirstName";
  cardMaps["phoneticFirstName"] = "PhoneticFirstName";
  cardMaps["lastName"] = "LastName";
  cardMaps["phoneticLastName"] = "PhoneticLastName";
  cardMaps["displayName"] = "DisplayName";
  cardMaps["nickName"] = "NickName";
  // Contact > Internet
  cardMaps["primaryEmail"] = "PrimaryEmail";
  cardMaps["secondEmail"] = "SecondEmail";
  cardMaps["aimScreenName"] = "_AimScreenName"; // NB: AIM.
	cardMaps["preferMailFormat"] = "PreferMailFormat";
	//cardMaps["allowRemoteContent"] = "AllowRemoteContent";
  // Contact > Phones
  cardMaps["workPhone"] = "WorkPhone";
  cardMaps["homePhone"] = "HomePhone";
  cardMaps["faxNumber"] = "FaxNumber";
  cardMaps["pagerNumber"] = "PagerNumber";
  cardMaps["cellularNumber"] = "CellularNumber";
  // Address > Home
  cardMaps["homeAddress"] = "HomeAddress";
  cardMaps["homeAddress2"] = "HomeAddress2";
  cardMaps["homeCity"] = "HomeCity";
  cardMaps["homeState"] = "HomeState";
  cardMaps["homeZipCode"] = "HomeZipCode";
  cardMaps["homeCountry"] = "HomeCountry";
  cardMaps["webPage2"] = "WebPage2";
  // Address > Work
  cardMaps["jobTitle"] = "JobTitle";
  cardMaps["department"] = "Department";
  cardMaps["company"] = "Company";
  cardMaps["workAddress"] = "WorkAddress";
  cardMaps["workAddress2"] = "WorkAddress2";
  cardMaps["workCity"] = "WorkCity";
  cardMaps["workState"] = "WorkState";
  cardMaps["workZipCode"] = "WorkZipCode";
  cardMaps["workCountry"] = "WorkCountry";
  cardMaps["webPage1"] = "WebPage1";
  // Other > (custom)
  cardMaps["custom1"] = "Custom1";
  cardMaps["custom2"] = "Custom2";
  cardMaps["custom3"] = "Custom3";
  cardMaps["custom4"] = "Custom4";
  // Other > Notes
  cardMaps["notes"] = "Notes";

	return cardMaps[propName] ? cardMaps[propName] : propName;
}

GCContactEntry.prototype.dump = function(card)
{
	var cardFields = [
				 // Contact > Name
         "FirstName",
         "PhoneticFirstName",
         "LastName",
         "PhoneticLastName",
         "DisplayName",
         "NickName",
          // Contact > Internet
         "PrimaryEmail",
         "SecondEmail",
         "_AimScreenName", // NB: AIM.
    		 "PreferMailFormat",
				 //"AllowRemoteContent",
    		 "allowRemoteContent",
          // Contact > Phones
         "WorkPhone",
         "HomePhone",
         "FaxNumber",
         "PagerNumber",
         "CellularNumber",
          // Address > Home
         "HomeAddress",
         "HomeAddress2",
         "HomeCity",
         "HomeState",
         "HomeZipCode",
         "HomeCountry",
         "WebPage2",
          // Address > Work
         "JobTitle",
         "Department",
         "Company",
         "WorkAddress",
         "WorkAddress2",
         "WorkCity",
         "WorkState",
         "WorkZipCode",
         "WorkCountry",
         "WebPage1",
         //Birthday
         "BirthYear",
         "BirthMonth",
         "BirthDay",
          // Other > (custom)
         "Custom1",
         "Custom2",
         "Custom3",
         "Custom4",
          // Other > Notes
         "Notes"
         ];
  var str = "";
  for (var i=0; i<cardFields.length; i++) {
		if (card.getProperty(cardFields[i], null)) str = str + card.getProperty(cardFields[i], null);
	}
	return str;
	//return card.convertToEscapedVCard()+card.phoneticLastName+card.phoneticFirstName+card.aimScreenName;
}

GCContactEntry.prototype.isChanged = function(card)
{
	var ret = this.dumpstr != this.dump(card);
	return ret;
}

GCContactEntry.prototype.getTitle = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return "";
	var gd = this.gd;
	var fullName = "";
	try {
		fullName = this.entry.gd::name.gd::fullName;
	} catch(e) {
	}
	return fullName.toString();
}
	
GCContactEntry.prototype.getPhotoLink = function(forAdd)
{

	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	var gd = this.gd;
	var linkXML = this.entry.atom::link.(@rel=="http://schemas.google.com/contacts/2008/rel#photo");
	var link = null;
	try {
		if (forAdd || (linkXML.@gd::etag).toString()) {
			link = linkXML.@href;
		}
	} catch(e) {
		dump(e);
	}

	return link;
}

GCContactEntry.prototype.getPhotoEtag = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	var gd = this.gd;
	var linkXML = this.entry.atom::link.(@rel=="http://schemas.google.com/contacts/2008/rel#photo");
	var etag = null;
	try {
		etag = (linkXML.@gd::etag).toString();
	} catch(e) {
		dump(e);
	}

	return etag;
}

GCContactEntry.prototype.getDispName = function()
{
	return this.getTitle();
}

GCContactEntry.prototype.getPrimaryAddress = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var gd = this.gd;
	var addr = "";
	try {
		//addr = this.entry.gd::email.(@primary=="true").@address;
		addr = this.entry.gd::email.(function::attribute("primary")=="true").@address;
	} catch(e) {
		addr = this.entry.gd::email.@address[0];
	}
	
	if (addr.toString() == "" && this.card) { //pseudo email
		addr = this.card.getProperty("PrimaryEmail", "");
	}
	
	return addr.toString();
}

GCContactEntry.prototype.getAddresses = function()
{
	//if (!this.entry) this.loadCache();
	var ret = new Array();
	if (!this.entry && !this.loadCache()) return ret;
	var gd = this.gd;
	var addresses = this.entry.gd::email;
	for (var i=0; i< addresses.length(); i++) {
		var address = addresses[i];
		var rel = address.@rel.toString();
		var relParts = rel.split("#");
		var addrObj = {
			address: address.@address.toString(),
			type: relParts[1]
		};
		try {
			addrObj.primary = address.@primary;
		} catch(e) {
		}
		ret.push(addrObj);
	}
	
	return ret;
}

GCContactEntry.prototype.print = function()
{
	dump("#####\n");
	dump("Id: " + this.getId()+"\n");
	var addrs = this.getAddresses();
	for (var i=0; i<addrs.length; i++) {
		dump("Address: " + addrs[i].address+"\n");
		dump("\tType: " + addrs[i].type+"\n");
		dump("\tPrimary: " + addrs[i].primary+"\n");
	}
}
